#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "fsubagentArray.h"

#define CHECK(a, b) assert(a == b)
#define CHECKNE(a, b) assert(a != b)
#define STRCHECK(a, b) assert(!strcmp(a, b))

static void saveSomeData(void)
{
	int i;
	int ival;
	const char *sval;
	char buf[64000];
	int buflen;
	int rownum;

	fsubagentArray *arr;
	fsubagentRow *row_1;	/* row oid suffix 1 */
	fsubagentRow *row_2;	/* row oid suffix 2 */
	fsubagentRow *rowx;
	fsubagentRowConstraints constraints;

	constraints.format = "iis";
	constraints.maxoid = strlen(constraints.format);

	arr = fsubagentArray_new(&constraints);
	CHECKNE(0, arr);
	CHECK(0, fsubagentArray_getRow(arr, 0));
	CHECK(0, fsubagentArray_getRow(arr, 1));

	row_1 = fsubagentArray_createRow(arr, &rownum);
	CHECKNE(0, row_1);
	CHECK(1, rownum);
	rowx = fsubagentArray_getRow(arr, 1);
	CHECK(rowx, row_1);

	row_2 = fsubagentArray_createRow(arr, &rownum);
	CHECKNE(0, row_2);
	CHECK(2, rownum);
	rowx = fsubagentArray_getRow(arr, 2);
	CHECK(rowx, row_2);
	CHECKNE(row_1, row_2);

	/* Check that initial values are zero or empty */
	ival = -1;
	CHECK(0, fsubagentRow_getInt(row_1, 1, &ival));
	CHECK(0, ival);
	sval = NULL;
	CHECK(0, fsubagentRow_getString(row_1, 3, &sval));
	CHECKNE(0, sval);
	CHECK(0, sval[0]);

	/* Now store values (for convenience, use the oid as the value, kinda) */
	CHECK(0, fsubagentRow_setInt(row_1, 1, 0x8811));
	CHECK(0, fsubagentRow_setInt(row_1, 2, 0x8812));
	CHECK(0, fsubagentRow_setString(row_1, 3, "8813"));
	CHECK(0, fsubagentRow_setInt(row_2, 1, 0x8821));
	CHECK(0, fsubagentRow_setInt(row_2, 2, 0x8822));
	CHECK(0, fsubagentRow_setString(row_2, 3, "8823"));

	/* Now retrieve values and make sure they're what we set earlier */
	CHECK(0, fsubagentRow_getInt(row_1, 1, &ival));
	CHECK(0x8811, ival);
	CHECK(0, fsubagentRow_getInt(row_1, 2, &ival));
	CHECK(0x8812, ival);
	CHECK(0, fsubagentRow_getString(row_1, 3, &sval));
	CHECK(0, strcmp(sval, "8813"));
	CHECK(0, fsubagentRow_getInt(row_2, 1, &ival));
	CHECK(0x8821, ival);
	CHECK(0, fsubagentRow_getInt(row_2, 2, &ival));
	CHECK(0x8822, ival);
	CHECK(0, fsubagentRow_getString(row_2, 3, &sval));
	CHECK(0, strcmp(sval, "8823"));

	/* Make sure we can find rows by content */
	CHECK(row_1, fsubagentArray_getRowByStr(arr, 3, "8813", &rownum));
	CHECK(rownum, 1);
	CHECK(row_2, fsubagentArray_getRowByStr(arr, 3, "8823", &rownum));
	CHECK(rownum, 2);
	CHECK(NULL,  fsubagentArray_getRowByStr(arr, 3, "blort", 0));

	/* Save to buffer */
	buflen = 0;
	CHECK(0, fsubagentArray_toBuf(arr, buf, sizeof(buf), &buflen));

	fsubagentArray_delete(arr);

	/* Load from buffer */
	arr = fsubagentArray_new(&constraints);
	CHECKNE(0, arr);
	row_1 = fsubagentArray_createRow(arr, NULL);
	CHECKNE(0, row_1);
	row_2 = fsubagentArray_createRow(arr, NULL);
	CHECKNE(0, row_2);

	printf("Info: buf = ");
	for (i=0; i<buflen; i++)
		printf("%02x ", 0xff & buf[i]);
	printf("\n");

	/* Now retrieve values and make sure they're what we set earlier */
	CHECK(0, fsubagentArray_fromBuf(arr, buf, buflen));

	/* Retrieve values and make sure they're what we set earlier */
	CHECK(0, fsubagentRow_getInt(row_1, 1, &ival));
	CHECK(0x8811, ival);
	CHECK(0, fsubagentRow_getInt(row_1, 2, &ival));
	CHECK(0x8812, ival);
	CHECK(0, fsubagentRow_getInt(row_2, 1, &ival));
	CHECK(0x8821, ival);
	CHECK(0, fsubagentRow_getInt(row_2, 2, &ival));
	CHECK(0x8822, ival);

	/* Make sure we can find rows by content */
	CHECK(row_1, fsubagentArray_getRowByStr(arr, 3, "8813", 0));
	CHECK(row_2, fsubagentArray_getRowByStr(arr, 3, "8823", 0));
	CHECK(NULL,  fsubagentArray_getRowByStr(arr, 3, "blort", 0));

	fsubagentArray_delete(arr);
}


int main(int argc, char **argv)
{
	(void) argc;
	(void) argv;
	saveSomeData();
	printf("fsubagentArray_test: pass\n");
	return 0;
}
