/* Program to dump a file saved by fsubagent_save() as ASCII.
 * FIXME: this is total code duplication, and will inevitably
 * be a maintenance nightmare.
 * Also, as of July 2001, this has not been code reviewed, so there 
 * are probably ugly bits lurking in it.
 */
#include "fsubagent.h"
#include "portable.h"
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>

#include "dprint.h"

#define MAXOIDLEN 80
#define fsubagentThing_MAGIC 0x5354

enum fsubagentThingType { 
	/* fsubagentRow: Row of scalars */
	SA_ROW = 'r', 
	/* fsubagentArray: Table with one integer index */
	SA_ARRAY = 'a',
	/* fsubagentArray2: Table with two integer indices */
	SA_ARRAY2 = 'b' 
};

int fsubagent_dump(const char*);
int fsubagent_dumpThing(const char*, int, int*);
int dumpRow(const char *, int, int*);
int dumpItem(const char*, int, int*);
int dumpArray(const char*, int, int*);
int dumpArray2(const char*, int, int*);
void die(void);

void die()
{
	printf("Aborted due to errors!\n");
	exit(1);
}

int main(int argc, char **argv)
{
	const char *filename;

	if (argc != 2) {
		printf("\
Usage: %s filename\n\
where filename names a file written by fsubagent_save()\n\
Note: treats all items of length 4 as integer, all others as strings\n",
			argv[0]);

		return EXIT_FAILURE;
	}
	
	filename = argv[1];

	if (fsubagent_dump(filename)) 
		die();
	

	return EXIT_SUCCESS;
}

int fsubagent_dump(const char *filename)
{
	off_t len;
	int n;
	int fd = open(filename, O_RDONLY);
	char *buf;

	/* new variables to prevent the need for a fsubagent object*/
	char *iobuf;
	int iobuflen = 0;
	/* need to allocate iobuf to something, because given code uses 'realloc'*/
	iobuf = malloc(128);

	DPRINT(("fsubagent_dump(): filename %s\n", filename));
	if (fd == -1) {
		DPRINT(("fsubagent_dump(): fail, can't open file %s\n", filename));
		return errno;
	}

	/* Get file length, do sanity check */
	len = lseek(fd, 0, SEEK_END);
	lseek(fd, 0, SEEK_SET);
	if (len > 1000000) {
		close(fd);
		DPRINT(("fsubagent_dump: fail at line %d\n", __LINE__));
		return E2BIG;
	}

	/* Make sure buffer's big enough */
	if (len > iobuflen) {
		buf = realloc(iobuf, len);
		if (!buf) {
			close(fd);
			DPRINT(("fsubagent_dump: fail at line %d\n", __LINE__));
			return ENOMEM;
		}
		iobuf = buf;
		iobuflen = len;
	}


	/* Read in the file */
	/* FIXME: should just memory-map it */
	n = read(fd, iobuf, len);
	close(fd);
	if (n != len) {
		DPRINT(("fsubagent_dump: fail at line %d\n", __LINE__));
		return EIO;
	}

	/* While there is more file left, load the next oid from disk,
	 * search for it in things, and load that thing from the file.
	 */
	buf = iobuf;
	while (len > 0) {
		int nused;
		int err = fsubagent_dumpThing(buf, len, &nused);
		if (err) {
			DPRINT(("fsubagent_dumpThing: fail at line %d\n", __LINE__));
			return err;
		}
		DPRINT(("fsubagent_dumpThing: thing at %d took %d\n", buf-iobuf, nused));
		buf += nused;
		len -= nused;
	}
	return 0;
}

int fsubagent_dumpThing(const char *buf, int buflen, int *bufused)
{
	/* format: 
	 * ThingRecord := magic(2) bodylen(2) ThingType(1) oidlen(1) oid(oidlen*2) ThingBody
	 * ThingBody := { RowBody | ArrayBody }
	 */
	
	int bodylen, magic;
	enum fsubagentThingType kind;
	int oidlen;
	int oid[MAXOIDLEN];
	int i;
	int origlen;

	origlen = buflen;

	DPRINT(("buf:"));
	i = buflen; 
	if (i > 32) i = 32;
	fsubagent_memdump(buf, i);
	DPRINT(("\n"));

	/* Read fixed-width part of item record */
	if (buflen < 6) {
		DPRINT(("fsubagent_dumpThing: fail at line %d\n", __LINE__));
		return EINVAL;
	}
	magic = BE_BYTES_TO_INT16v(buf);
	buf += 2;
	buflen -= 2;
	if (magic != fsubagentThing_MAGIC) {
		DPRINT(("fsubagent_dumpThing: fail at line %d\n", __LINE__));
		return EINVAL;
	}

	bodylen = BE_BYTES_TO_INT16v(buf);
	buf += 2;
	buflen -= 2;
	kind = 0xff & *buf++;
	buflen--;
	oidlen = 0xff & *buf++;
	buflen--;

	if (bodylen < 1 || bodylen > buflen) {
		DPRINT(("fsubagent_dumpThing: fail at line %d\n", __LINE__));
		return EINVAL;
	}

	if (oidlen < 1 || oidlen >= MAXOIDLEN) {
		DPRINT(("fsubagent_dumpThing: fail at line %d\n", __LINE__));
		return EINVAL;
	}

	printf("OID: ");
	/* Read in variable-length portion */
	oid[0] = BE_BYTES_TO_INT16v(buf);
	buf+=2;
	buflen-=2;
	printf("%d", oid[0]);
	for (i=1; i<oidlen; i++, buf += 2, buflen -= 2) {
		oid[i] = BE_BYTES_TO_INT16v(buf);
		printf(".%d", oid[i]);
	}
	oid[i] = fsubagent_END;
	
	printf("    BODYLEN: %d     THINGTYPE: ", bodylen);

	if (kind == SA_ROW) {
		printf("SA_ROW\n"); 
		if (dumpRow(buf, bodylen, oid)) 
			die();
	}
	else if (kind == SA_ARRAY) {
		printf("SA_ARRAY\n"); 
		if (dumpArray(buf, bodylen, oid)) 
			die();
	}
	else if (kind == SA_ARRAY2) {
		printf("SA_ARRAY2\n");
		if (dumpArray2(buf, bodylen, oid)) 
			die();
	}
	else
		printf("INVALID\n"); 

	/*buf += bodylen;*/
	buflen -= bodylen;

	*bufused = origlen - buflen;

	DPRINT(("fsubagent_dumpThing: done\n"));
	return 0;
}

int dumpRow(const char *buf, int buflen, int *oid)
{
	int retval = 0;
	int i;
	int oldoidlimit;

	/* format: 
	 * RowBody := ItemRecord+
	 * ItemRecord := magic(2) len(2) index(2) ItemBody
	 */

	while (buflen > 0) {
		int len, index, magic;
		int err;

		/* Read fixed-width part of item record */
		if (buflen < 6) {
			retval = EINVAL;
			break;
		}
		magic = BE_BYTES_TO_INT16v(buf);
		buf += 2;
		buflen -= 2;
		if (magic != fsubagentRow_MAGIC)
			return EINVAL;

		len = BE_BYTES_TO_INT16v(buf);
		buf += 2;
		index = BE_BYTES_TO_INT16v(buf);
		buf += 2;
		buflen -= 4;

		if (len < 1 || len > buflen) {
			retval = EINVAL;
			break;
		}

		if (index < 0) {
			retval = EINVAL;
			break;
		}

		i = 0;

		while (oid[i] != fsubagent_END) {
			i++;
		}

		oldoidlimit = i;

		oid[i++] = index+1;
		oid[i] = fsubagent_END;
	
		/* Read in variable-length portion */
		err = dumpItem(buf, len, oid);
		if (err) {
			retval = err;
			break;
		}
		buf += len;
		buflen -= len;

		oid[oldoidlimit] = fsubagent_END;
	}
	printf("\n");

	return retval;
}

int dumpItem(const char *buf, int buflen, int *oid)
{
	/* format: value(buflen)
	 * Type is implicit; it's stored in fsubagentItemRow.constraints->format[]
	 * If type is fsubagentItem_STRING, buf and buflen do NOT include 
	 * nul terminating byte; terminating nul is appended here.
	 * Byte order is big endian.
	 */
	int i;
	enum fsubagentItemType kind;

	/* KLUDGE: guess type */
	if (buflen == sizeof(int)) {
		kind = fsubagentItem_INT;
	} else {
		kind = fsubagentItem_STRING;
	}

	i = 0;
	printf("%d", oid[i++]);

	while (oid[i] != fsubagent_END) {
		printf(".%d", oid[i++]);
	}

	printf(" ");

	switch (kind) {
	case fsubagentItem_INT:
		if (buflen != sizeof(int)) {
			/* bad format */
			return EINVAL;
		}
	    printf("%9d", BE_BYTES_TO_INT32v(buf));
		break;
	case fsubagentItem_STRING:
		fwrite(buf, 1, buflen, stdout);
		break;
	default:
		return EINVAL;
	}

	printf("\n");

	return 0;
}

int dumpArray(const char *buf, int buflen, int* oid)
{
	int retval = 0;
	int i;
	int oldoidlimit;

	/* format: 
	 * ArrayBody := RowRecord+
	 * RowRecord := magic(2) len(2) index(2) RowBody
	 * Note: index is 1 based
	 */

	while (buflen > 0) {
		int magic, len, index;
		int err;

		DPRINT(("fsubagentArray_fromBuf: buflen at top: %d\n", buflen));

		/* Read fixed-width part of item record */
		if (buflen < 6) {
			retval = EINVAL;
			DPRINT(("fromBuf failed, line %d\n", __LINE__));
			break;
		}
		magic = BE_BYTES_TO_INT16v(buf);
		buf += 2;
		buflen -= 2;
		if (magic != fsubagentArray_MAGIC) {
			DPRINT(("fromBuf failed, line %d\n", __LINE__));
			return EINVAL;
		}
		len = BE_BYTES_TO_INT16v(buf);
		buf += 2;
		buflen -= 2;
		index = BE_BYTES_TO_INT16v(buf);
		buf += 2;
		buflen -= 2;

		if (len < 1 || len > buflen) {
			retval = EINVAL;
			DPRINT(("fromBuf failed, line %d\n", __LINE__));
			break;
		}

		if (index < 1) {
			retval = EINVAL;
			DPRINT(("fromBuf failed, line %d\n", __LINE__));
			break;
		}

		/*	row = fsubagentArray_getRow(this, index);
			if (!row) {
				// FIXME: this is a bit fragile
				int rownum;
				row = fsubagentArray_createRow(this, &rownum);
				assert(index == rownum);
			}
			if (!row) {
				retval = ENOMEM;
				DPRINT(("fromBuf failed, line %d\n", __LINE__));
				break;
			}
		*/
		i = 0;

		while (oid[i] != fsubagent_END) {
			i++;
		}

		oldoidlimit = i;

		oid[i++] = index+1;
		oid[i] = fsubagent_END;


		/* Read in variable-length portion */
		err = dumpRow(buf, len, oid);
		if (err) {
			retval = err;
			DPRINT(("fromBuf failed, line %d\n", __LINE__));
			break;
		}
		buf += len;
		buflen -= len;
		DPRINT(("fsubagentArray_fromBuf: buflen at bottom: %d\n", buflen));

		oid[oldoidlimit] = fsubagent_END;
	}

	return retval;
}

int dumpArray2(const char *buf, int buflen, int *oid)
{
	int retval = 0;
	int i;
	int oldoidlimit;

	/* format: 
	 * ArrayBody := RowRecord+
	 * RowRecord := magic(2) len(2) index(2) RowBody
	 * Note: index is 1 based
	 */

	while (buflen > 0) {
		int magic, len, index;
		int err;

		DPRINT(("fsubagentArray2_fromBuf: at top: buflen %d\n", buflen));
		DPRINT(("buf:"));
		len = buflen; 
		if (len > 32) len = 32;
		fsubagent_memdump(buf, len);
		DPRINT(("\n"));

		/* Read fixed-width part of item record */
		if (buflen < 6) {
			retval = EINVAL;
			DPRINT(("fromBuf failed, file %s, line %d\n", __FILE__, __LINE__));
			break;
		}
		magic = BE_BYTES_TO_INT16v(buf);
		buf += 2;
		buflen -= 2;
		if (magic != fsubagentArray2_MAGIC) {
			DPRINT(("fromBuf failed, file %s, line %d\n", __FILE__, __LINE__));
			return EINVAL;
		}
		len = BE_BYTES_TO_INT16v(buf);
		buf += 2;
		buflen -= 2;
		index = BE_BYTES_TO_INT16v(buf);
		buf += 2;
		buflen -= 2;

		// We can get rid of this 'if' once we make toBuf not output zero-length records!
		if (len == 0)
			continue;

		if (len < 1 || len > buflen) {
			retval = EINVAL;
			DPRINT(("fromBuf failed; len %d not in range 1..%d, file %s, line %d\n", 
				len, buflen, __FILE__, __LINE__));
			break;
		}

		if (index < 1) {
			retval = EINVAL;
			DPRINT(("fromBuf failed, file %s, line %d\n", __FILE__, __LINE__));
			break;
		}

		DPRINT(("fsubagentArray2_fromBuf: looking up array at index %d\n", index));
		/*
		array = fsubagentArray2_getArray(this, index);
		if (!array) {
			// FIXME: this is a bit fragile
			int arrnum;
			DPRINT(("fsubagentArray2_fromBuf: creating array at index %d\n", index));
			array = fsubagentArray2_createArray(this, &arrnum);
			assert(index == arrnum);
		}
		if (!array) {
			retval = ENOMEM;
			DPRINT(("fromBuf failed, file %s, line %d\n", __FILE__, __LINE__));
			break;
		}
		*/
		i = 0;

		while (oid[i] != fsubagent_END) {
			i++;
		}

		oldoidlimit = i;

		oid[i++] = index+1;
		oid[i] = fsubagent_END;
		
		/* Read in variable-length portion */
		err = dumpArray(buf, len, oid);
		if (err) {
			retval = err;
			DPRINT(("fromBuf failed, file %s, line %d\n", __FILE__, __LINE__));
			break;
		}
		buf += len;
		buflen -= len;

		DPRINT(("fsubagentArray2_fromBuf: at bottom: buflen %d\n", buflen));

		oid[oldoidlimit] = fsubagent_END;
	}

	return retval;
}

