#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "fsubagentRow.h"

#define CHECK(a, b) assert(a == b)
#define CHECKNE(a, b) assert(a != b)
#define STRCHECK(a, b) assert(!strcmp(a, b))

static void saveSomeData(void)
{
	int ival;
	const char *sval;
	fsubagentRow *row;
	fsubagentRowConstraints constraints;
	char buf[64000];
	int buflen;
	int i;

	constraints.format = "sii";
	constraints.maxoid = strlen(constraints.format);

	row = fsubagentRow_new(&constraints);
	CHECKNE(row, NULL);

	/* Make sure out-of-bounds indices are caught */
	CHECKNE(0, fsubagentRow_setInt(row, 4, 0));

	/* Store some values that will be overwritten */
	CHECK(0, fsubagentRow_setInt(row, 2, -1));

	/* Now store values (for convenience, use the oid as the value, kinda) */
	CHECK(0, fsubagentRow_setString(row, 1, "31"));
	CHECK(0, fsubagentRow_setInt(row, 2, 32));
	CHECK(0, fsubagentRow_setInt(row, 3, 33));

	/* Check type checking */
	CHECKNE(0, fsubagentRow_setString(row, 2, "spam"));
	CHECKNE(0, fsubagentRow_getString(row, 2, &sval));

	/* Now retrieve values and make sure they're what we set earlier */

	CHECK(0, fsubagentRow_getString(row, 1, &sval));
	STRCHECK(sval, "31");
	CHECK(0, fsubagentRow_getInt(row, 2, &ival));
	CHECK(32, ival);
	CHECK(0, fsubagentRow_getInt(row, 3, &ival));
	CHECK(33, ival);

	/* Save to buffer */
	buflen = 0;
	CHECK(0, fsubagentRow_toBuf(row, buf, sizeof(buf), &buflen));

	fsubagentRow_delete(row);

	/* Load from buffer */
	row = fsubagentRow_new(&constraints);
	CHECKNE(row, NULL);

	printf("Info: buf = ");
	for (i=0; i<buflen; i++)
		printf("%02x ", 0xff & buf[i]);
	printf("\n");

	/* Now retrieve values and make sure they're what we set earlier */
	CHECK(0, fsubagentRow_fromBuf(row, buf, buflen));

	CHECK(0, fsubagentRow_getString(row, 1, &sval));
	STRCHECK(sval, "31");
	CHECK(0, fsubagentRow_getInt(row, 2, &ival));
	CHECK(32, ival);
	CHECK(0, fsubagentRow_getInt(row, 3, &ival));
	CHECK(33, ival);

	/* One more time, to make sure we can load twice */
	CHECK(0, fsubagentRow_fromBuf(row, buf, buflen));

	CHECK(0, fsubagentRow_getString(row, 1, &sval));
	STRCHECK(sval, "31");
	CHECK(0, fsubagentRow_getInt(row, 2, &ival));
	CHECK(32, ival);
	CHECK(0, fsubagentRow_getInt(row, 3, &ival));
	CHECK(33, ival);

	fsubagentRow_delete(row);
}

int main(int argc, char **argv)
{
	(void) argc;
	(void) argv;
	saveSomeData();
	printf("fsubagentRow_test: pass\n");
	return 0;
}
