#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "fsubagent.h"

#define CHECK(a, b) assert(a == b)
#define CHECKNE(a, b) assert(a != b)
#define STRCHECK(a, b) assert(!strcmp(a, b))

static void saveSomeData(void)
{
	static int myoid[16];
	int ival;
	const char *sval;
	fsubagent *sa = fsubagent_new();

	fsubagentArray *arr1;	/* one-index table oid {1,*,*} */
	fsubagentRow *row1_1;	/*   oid {1,*,1}               */
	fsubagentArray *arr22;	/* one-index table oid 2,*,*}  */
	fsubagentRow *row22_1;	/*   oid {2,*,1}               */
	fsubagentRow *row22_2;	/*   oid {2,*,2}               */
	fsubagentRow *row3;		/* scalar row oid {3,*,0}      */
	fsubagentRow *rowx;

	fsubagentArray2 *root;	/* two-index table oid {4,*,*,*} */
	fsubagentArray *root1;	/*   oid {4,*,1,*}               */
	fsubagentRow *root1r1;	/*     oid {4,*,1,1}             */
	fsubagentArray *root2;	/*   oid {4,*,2,*}               */
	fsubagentRow *root2r1;	/*     oid {4,*,2,1}             */

	CHECKNE(sa, NULL);
	CHECK(0, fsubagent_setFilename(sa, "fsubagent_test.dat"));
	myoid[0] = 1; 
	myoid[0] = 1; 

	/* Create one array at {1} maxindex 1 */
	myoid[0] = 1; myoid[1] = fsubagent_END;
	arr1 = fsubagent_createArray(sa, myoid, "i");
	CHECKNE(0, arr1);
	CHECK(0, fsubagentArray_getRow(arr1, 0));
	CHECK(0, fsubagentArray_getRow(arr1, 1));
	row1_1 = fsubagentArray_createRow(arr1, NULL);
	CHECKNE(0, row1_1);
	rowx = fsubagentArray_getRow(arr1, 1);
	CHECK(row1_1, rowx);

	/* Create a second array at {2, 2} maxindex 2 */
	myoid[0] = 2; myoid[1] = 2; myoid[2] = fsubagent_END;
	arr22 = fsubagent_createArray(sa, myoid, "ii");
	CHECKNE(0, arr22);
	CHECK(0, fsubagentArray_getRow(arr22, 0));
	CHECK(0, fsubagentArray_getRow(arr22, 1));

	row22_1 = fsubagentArray_createRow(arr22, NULL);
	CHECKNE(0, row22_1);
	rowx = fsubagentArray_getRow(arr22, 1);
	CHECK(rowx, row22_1);

	/* Store some values that will be overwritten */
	CHECK(0, fsubagentRow_setInt(row22_1, 1, -1));

	row22_2 = fsubagentArray_createRow(arr22, NULL);
	CHECKNE(0, row22_2);
	rowx = fsubagentArray_getRow(arr22, 2);
	CHECK(rowx, row22_2);
	CHECKNE(row22_1, row22_2);

	/* Create a scalar row at {3} maxindex 3 */
	myoid[0] = 3; myoid[1] = fsubagent_END;
	row3 = fsubagent_createScalarRow(sa, myoid, "sii");
	CHECKNE(0, row3);
	rowx = fsubagent_createScalarRow(sa, myoid, "sii");
	CHECK(NULL, rowx);

	/* Create a two-index table at {4}; rows have one int. */
	myoid[0] = 4; myoid[1] = fsubagent_END;
	root = fsubagent_createArray2(sa, myoid, "i");
	CHECKNE(0, root);
	/* Create two subarrays each with one row. */
	root1 = fsubagentArray2_createArray(root, NULL);
	CHECKNE(0, root1);
	CHECK(fsubagentArray2_getArray(root, 1), root1);
	root1r1 = fsubagentArray_createRow(root1, NULL);
	CHECKNE(0, root1r1);
	root2 = fsubagentArray2_createArray(root, NULL);
	CHECKNE(0, root2);
	CHECK(fsubagentArray2_getArray(root, 2), root2);
	root2r1 = fsubagentArray_createRow(root2, NULL);
	CHECKNE(0, root2r1);

	CHECK(0, fsubagentRow_setInt(root1r1, 1, 0x4111));
	CHECK(0, fsubagentRow_setInt(root2r1, 1, 0x4211));

	/* Make sure out-of-bounds indices are caught */
	CHECKNE(0, fsubagentRow_setInt(row1_1, 2, 0));

	/* Store some values that will be overwritten */
	CHECK(0, fsubagentRow_setInt(row22_2, 1, -1));

	/* Now store values (for convenience, use the oid as the value, kinda) */
	CHECK(0, fsubagentRow_setInt(row1_1, 1, 0));
	CHECK(0, fsubagentRow_addInt(row1_1, 1, 111));
	CHECK(0, fsubagentRow_setInt(row22_1, 1, 2211));
	CHECK(0, fsubagentRow_setInt(row22_1, 2, 2212));
	CHECK(0, fsubagentRow_setInt(row22_2, 1, 2221));
	CHECK(0, fsubagentRow_setInt(row22_2, 2, 2222));
	CHECK(0, fsubagentRow_setString(row3, 1, "31"));
	CHECK(0, fsubagentRow_setInt(row3, 2, 32));
	CHECK(0, fsubagentRow_setInt(row3, 3, 33));

	/* Check type checking */
	CHECKNE(0, fsubagentRow_setString(row3, 2, "spam"));
	CHECKNE(0, fsubagentRow_getString(row3, 2, &sval));

	/* Now retrieve values and make sure they're what we set earlier */
	CHECK(0, fsubagentRow_getInt(row1_1, 1, &ival));
	CHECK(111, ival);
	CHECK(0, fsubagentRow_getInt(row22_1, 1, &ival));
	CHECK(2211, ival);
	CHECK(0, fsubagentRow_getInt(row22_1, 2, &ival));
	CHECK(2212, ival);
	CHECK(0, fsubagentRow_getInt(row22_2, 1, &ival));
	CHECK(2221, ival);
	CHECK(0, fsubagentRow_getInt(row22_2, 2, &ival));
	CHECK(2222, ival);

	CHECK(0, fsubagentRow_getString(row3, 1, &sval));
	STRCHECK(sval, "31");
	CHECK(0, fsubagentRow_getInt(row3, 2, &ival));
	CHECK(32, ival);
	CHECK(0, fsubagentRow_getInt(row3, 3, &ival));
	CHECK(33, ival);

	CHECK(0, fsubagentRow_getInt(root1r1, 1, &ival));
	CHECK(ival, 0x4111);
	CHECK(0, fsubagentRow_getInt(root2r1, 1, &ival));
	CHECK(ival, 0x4211);

	/* Save to disk */
	CHECK(0, fsubagent_save(sa));

	fsubagent_delete(sa);
}

static void loadAndCheckData(void)
{
	static int myoid[16];
	int ival;
	const char *sval;
	int oid[16];
	fsubagent *sa = fsubagent_new();

	fsubagentArray *arr1;	/* one-index table oid {1,*,*} */
	fsubagentRow *row1_1;	/*   oid {1,*,1}               */
	fsubagentArray *arr22;	/* one-index table oid 2,*,*}  */
	fsubagentRow *row22_1;	/*   oid {2,*,1}               */
	fsubagentRow *row22_2;	/*   oid {2,*,2}               */
	fsubagentRow *row3;		/* scalar row oid {3,*,0}      */

	fsubagentArray2 *root;	/* two-index table oid {4,*,*,*} */
	fsubagentArray *root1;	/*   oid {4,*,1,*}               */
	fsubagentRow *root1r1;	/*     oid {4,*,1,1}             */
	fsubagentArray *root2;	/*   oid {4,*,2,*}               */
	fsubagentRow *root2r1;	/*     oid {4,*,2,1}             */

	CHECKNE(sa, NULL);
	CHECK(0, fsubagent_setFilename(sa, "fsubagent_test.dat"));

	oid[0] = 1; oid[1] = fsubagent_END;
	arr1 = fsubagent_createArray(sa, oid, "i");

	oid[0] = 2; oid[1] = 2; oid[2] = fsubagent_END;
	arr22 = fsubagent_createArray(sa, oid, "ii");

	oid[0] = 3; oid[1] = fsubagent_END;
	row3 = fsubagent_createScalarRow(sa, oid, "sii");
	CHECKNE(row3, 0);

	/* Create a two-index table at {4}; rows have one int. */
	myoid[0] = 4; myoid[1] = fsubagent_END;
	root = fsubagent_createArray2(sa, myoid, "i");
	CHECKNE(0, root);

	/* Load from disk */
	CHECK(0, fsubagent_load(sa));

	/* Now retrieve values and make sure they're what we set earlier */
	row1_1 = fsubagentArray_getRow(arr1, 1);
	CHECKNE(0, row1_1);
	row22_1 = fsubagentArray_getRow(arr22, 1);
	CHECKNE(0, row22_1);
	row22_2 = fsubagentArray_getRow(arr22, 2);
	CHECKNE(0, row22_2);

	CHECK(0, fsubagentRow_getInt(row1_1, 1, &ival));
	CHECK(111, ival);
	CHECK(0, fsubagentRow_getInt(row22_1, 1, &ival));
	CHECK(2211, ival);
	CHECK(0, fsubagentRow_getInt(row22_1, 2, &ival));
	CHECK(2212, ival);
	CHECK(0, fsubagentRow_getInt(row22_2, 1, &ival));
	CHECK(2221, ival);
	CHECK(0, fsubagentRow_getInt(row22_2, 2, &ival));
	CHECK(2222, ival);

	CHECK(0, fsubagentRow_getString(row3, 1, &sval));
	STRCHECK(sval, "31");
	CHECK(0, fsubagentRow_getInt(row3, 2, &ival));
	CHECK(32, ival);
	CHECK(0, fsubagentRow_getInt(row3, 3, &ival));
	CHECK(33, ival);

	root1 = fsubagentArray2_getArray(root, 1);
	CHECKNE(0, root1);
	root1r1 = fsubagentArray_getRow(root1, 1);
	CHECKNE(0, root1r1);
	root2 = fsubagentArray2_getArray(root, 2);
	CHECKNE(0, root2);
	root2r1 = fsubagentArray_getRow(root2, 1);
	CHECKNE(0, root2r1);

	CHECK(0, fsubagentRow_getInt(root1r1, 1, &ival));
	CHECK(ival, 0x4111);
	CHECK(0, fsubagentRow_getInt(root2r1, 1, &ival));
	CHECK(ival, 0x4211);

	fsubagent_delete(sa);
}

int main(int argc, char **argv)
{
	(void) argc;
	(void) argv;
	saveSomeData();
	loadAndCheckData();
	printf("fsubagent_test: pass\n");
	return 0;
}
